# This source code is part of the Gecos package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__author__ = "Patrick Kunzmann"
__all__ = ["rgb_to_lab", "lab_to_rgb"]

import warnings
import numpy as np
from skimage.color import rgb2lab, lab2rgb


def rgb_to_lab(rgb):
    """
    Convert one or multiple *RGB* color(s) to *Lab* color(s).
    
    Parameters
    ----------
    rgb : array-like, shape=(..., 3)
        The color to be converted.
        The only restriction on the array shape is that the length of
        the last dimension must be 3.
        *RGB* values are in range (0,1).
    
    Returns
    -------
    lab : ndarray
        The converted color.
    """
    return _convert_color(rgb, rgb2lab)

def lab_to_rgb(lab):
    """
    Convert one or multiple *Lab* color(s) to *RGB* color(s).
    
    Parameters
    ----------
    lab : array-like, shape=(..., 3)
        The color to be converted.
        The only restriction on the array shape is that the length of
        the last dimension must be 3.
    
    Returns
    -------
    rgb : ndarray
        The converted color. *RGB* values are in range (0,1).
    """
    return _convert_color(lab, lab2rgb)

def _convert_color(color, convert_func):
    if not isinstance(color, np.ndarray):
        color = np.array(color, dtype=float)
    else:
        color = color.astype(float, copy=False)
    original_shape = color.shape
    # Temporarily reshape into a scikit-image-accepted shape
    color = color.reshape([1,-1,3])
    # Convert color(s) using scikit-image
    # Ignore warnings about colors being not convertible
    # - they will be set later to NaN anyway
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        conv_color = convert_func(color)
    # Color components that cannot be properly converted are clamped
    # to (0/1) -> Use this to detect invalid values
    conv_color[
        :,
        (conv_color[0] == 0).any(axis=-1) | (conv_color[0] == 1).any(axis=-1),
        :
    ] = np.nan
    conv_color = conv_color.reshape(original_shape)
    return conv_color


from multiprocessing import Pool
from os.path import join, dirname, realpath, isfile
import itertools
import argparse
import copy
import sys
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import ListedColormap
import biotite.sequence as seq
import biotite.sequence.align as align
import biotite.sequence.io.fasta as fasta
import biotite.sequence.graphics as graphics
from .space import ColorSpace
from .optimizer import ColorOptimizer, DefaultScoreFunction
from .file import write_color_scheme


FIGURE_WIDTH = 8.0
EXAMPLE_FILE_NAME \
    = join(dirname(realpath(__file__)), "example_alignment.fasta")


def handle_error(func):
    def wrapped(*args, **kwargs):
        try:
            func(*args, **kwargs)
        except InputError as e:
            print(e, file=sys.stderr)
            sys.exit(1)
        except Exception as e:
            print("An unexpected error occured:\n", file=sys.stderr)
            raise
    return wrapped


class InputError(Exception):
    pass


def _optimize(seed, alphabet, score_func, space, constraints, 
              nsteps, beta_start, beta_end, stepsize_start, stepsize_end):
    """
    Worker function used for parallel execution of simulated annealing.
    """
    np.random.seed(seed)
    optimizer = ColorOptimizer(
        alphabet, score_func, space, constraints
    )
    optimizer.optimize(
        nsteps, beta_start, beta_end, stepsize_start, stepsize_end
    )
    return optimizer.get_result()

@handle_error
def main(args=None, result_container=None, show_plots=True):
    parser = argparse.ArgumentParser(
        description="This program automatically generates a color scheme for "
                    "sequence alignments. "
                    "The perceptual color differences correspond to the "
                    "distances in a given substitution matrix."
                    "\n"
                    "The algorithm tries to find an optimal color scheme by "
                    "means of color differences by performing a "
                    "Metropolis-Monte-Carlo optimization in color space.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    
    space_group  = parser.add_argument_group(
        title="Color space",
        description=None
    )
    matrix_group = parser.add_argument_group(
        title="Substitution matrix and alphabet",
        description=None
    )
    opt_group = parser.add_argument_group(
        title="Optimization",
        description=None
    )
    output_group = parser.add_argument_group(
        title="Output files",
        description=None
    )
    vis_group    = parser.add_argument_group(
        title="Visualization",
        description=None
    )

    space_group.add_argument(
        "--dry-run", "-n", action="store_true",
        help="Show only the customized color space and terminate the program."
    )
    space_group.add_argument(
        "--smin", type=int,
        help="All colors in the space must "
            "have the specified saturation at minimum "
            "(a^2 + b^2 >= smin^2).",
        metavar="S"
    )
    space_group.add_argument(
        "--smax", type=int,
        help="All colors in the space must "
             "have the specified saturation at maximum "
             "(a^2 + b^2 <= smax^2).",
        metavar="S"
    )
    space_group.add_argument(
        "--lmin", type=int,
        help="All colors in the space must "
             "have the specified lightness at minimum.",
        metavar="L"
    )
    space_group.add_argument(
        "--lmax", type=int,
        help="All colors in the space must "
             "have the specified lightness at maximum.",
        metavar="L"
    )
    space_group.add_argument(
        "--amin", type=int,
        help="All colors in the space must "
             "have the specified 'a*' value at minimum.",
        metavar="A"
    )
    space_group.add_argument(
        "--amax", type=int,
        help="All colors in the space must "
             "have the specified 'a*' value at maximum.",
        metavar="A"
    )
    space_group.add_argument(
        "--bmin", type=int,
        help="All colors in the space must "
             "have the specified 'b*' value at minimum.",
        metavar="B"
    )
    space_group.add_argument(
        "--bmax", type=int,
        help="All colors in the space must "
            "have the specified 'b*' value at maximum.",
        metavar="B"
    )

    matrix_group.add_argument(
        "--alphabet", "-a",
        help="A custom alphabet to generate the scheme for (e.g. 'ACGT'). "
             "By default an alphabet containing the 20 amino acids is used. ",
        metavar="SYMBOLS"
    )
    matrix_group.add_argument(
        "--matrix", "-m", default="BLOSUM62",
        help="The substitution matrix to calculate the pairwise symbol "
             "distances from. "
             "Can be an NCBI substitution matrix name (e.g. 'BLOSUM62') or "
             "alternatively a substitution matrix file in NCBI format "
             "(e.g. 'blosum62.mat').",
        metavar="MATRIX"
    )

    opt_group.add_argument(
        "--contrast", default=700, type=int,
        help="The contrast factor controls how strongly the symbols are "
             "pushed to the edges of the color space. "
             "At the minimum value '0' contrast is not rewarded.",
        metavar="FACTOR"
    )
    opt_group.add_argument(
        "--constraint", "-c", nargs=4, action="append",
        help="Constrain a symbol to a fixed position in color space. "
             "This argument takes four values: a symbol, "
             "and the color channels 'l*', 'a*' and 'b*'"
             "(e.g. 'A 65 10 15')."
             "Can be repeated to add constraints for multiple symbols.",
        metavar=("SYMBOL", "L", "A", "B")
    )
    opt_group.add_argument(
        "--nsteps", default=20000, type=int,
        help="The optimization process performs simulated annealing in order "
             "to find the optimal conformation of the color scheme. "
             "This parameter sets the total amount of optimization steps. "
             "With a higher number of steps the quality of the optimization "
             "increases at the cost of a longer runtime.",
        metavar="NUMBER"
    )
    opt_group.add_argument(
        "--delta", default="CIEDE2000",
        choices=["CIE76", "CIEDE94", "CIEDE2000"],
        help="The formula to use for the calculation of perceptual "
             "differences.",
        metavar="FORMULA"
    )
    opt_group.add_argument(
        "--seed", type=float,
        help="Start seed used for seeding the parallel runs. "
             "By default the seed is chosen randomly.",
        metavar="NUMBER"
    )
    opt_group.add_argument(
        "--nruns", default=16, type=int,
        help="Number of optimizations to run. "
             "From these runs, the color scheme with the best score is "
             "selected.",
        metavar="NUMBER"
    )
    opt_group.add_argument(
        "--nthreads", type=int,
        help="Number of optimization runs to run in parallel. "
             "By default this is equal to the number of CPUs.",
        metavar="NUMBER"
    )
    opt_group.add_argument(
        "--beta-start", default=1, type=float,
        help="Inverse temperature at start of simulated annealing.",
        metavar="FLOAT",
    )
    opt_group.add_argument(
        "--beta-end", default=500, type=float,
        help="Inverse temperature at end of simulated annealing.",
        metavar="FLOAT",
    )   
    opt_group.add_argument(
        "--stepsize-start", default=10, type=float,
        help="Step size temperature at start of simulated annealing.",
        metavar="FLOAT",
    )
    opt_group.add_argument(
        "--stepsize-end", default=0.2, type=float,
        help="Step size temperature at end of simulated annealing.",
        metavar="FLOAT",
    )     
    
    output_group.add_argument(
        "--scheme-file", "-f",
        type=argparse.FileType(mode="w"), default=sys.stdout,
        help="Write the generated color scheme into the specified file. "
             "The scheme is saved as Biotite-compatible JSON file. "
             "By default the scheme is output to STDOUT.",
        metavar="FILE"
    )
    output_group.add_argument(
        "--name", default="scheme",
        help="The name of the color scheme that is used in the JSON file.",
        metavar="NAME"
    )
    output_group.add_argument(
        "--score-file", type=argparse.FileType(mode="w"),
        help="Write the scores during the color scheme optimization "
             "into the specified file. "
             "Each line corresponds to one optimization step.",
        metavar="FILE"
    )

    vis_group.add_argument(
        "--lightness", type=int,
        help="Set the lightness level for color space visualization. "
             "The plots generated via the '--show-space' and the "
             "'--show-scheme' options show the color space excerpt at the "
             "given lightness level. "
             "By default the mean of the minimum and maximum lightness of the "
             "space is chosen.",
        metavar="LIGHTNESS"
    )
    vis_group.add_argument(
        "--show-space", action="store_true",
        help="Show the generated color space. "
             "The lightness of the '--lightness' option is used here. "
             "The gray area cannot be displayed in RGB space at the given "
             "lightness."
    )
    vis_group.add_argument(
        "--show-scheme", action="store_true",
        help="Show the distribution of alphabet symbols in the color space."
             "The lightness of the '--lightness' option is used here. "
             "The gray area cannot be displayed in RGB space at the given "
             "lightness."
    )
    vis_group.add_argument(
        "--show-example", action="store_true",
        help="Show an example multiple sequence alignment "
             "with newly generated color space. "
             "Cannot be used in combination with a custom '--alphabet' value."
    )
    vis_group.add_argument(
        "--show-score", action="store_true",
        help="Show a plot of the score during the optimization process."
    )

    args = parser.parse_args(args=args)


    alphabet = parse_alphabet(args.alphabet)
    matrix = parse_matrix(args.matrix, alphabet)
    
    # This lightness is only used fotr visualization purposes
    if args.lightness is not None:
        lightness = args.lightness
    elif args.lmin is not None and args.lmax is not None:
        lightness = (args.lmin + args.lmax) // 2
    else:
        lightness = 50

    space = ColorSpace()
    adjust_saturation(space, args.smin, args.smax)
    adjust_l(space, args.lmin, args.lmax)
    adjust_a(space, args.amin, args.amax)
    adjust_b(space, args.bmin, args.bmax)
    if args.dry_run:
        figure = plt.figure(figsize=(FIGURE_WIDTH, FIGURE_WIDTH))
        ax = figure.gca()
        show_space(ax, space, lightness)
        figure.tight_layout()
        if show_plots:
            plt.show()
        return

    constraints = np.full((len(alphabet), 3), np.nan)
    if args.constraint is not None:
        for symbol, l, a, b in args.constraint:
            constraints[alphabet.encode(symbol)] = (l,a,b)
    
    score_func = DefaultScoreFunction(matrix, args.contrast, args.delta)   
    
    # Simulated annealing     
    if args.seed is not None:
        np.random.seed(int(args.seed))
    # Different random seed for each run
    seeds = np.random.randint(0, 1000000, size=args.nruns)

    with Pool(args.nthreads) as p:
        results = p.starmap(_optimize, zip(
            seeds,
            itertools.repeat(matrix.get_alphabet1()),
            itertools.repeat(score_func),
            itertools.repeat(space),
            itertools.repeat(constraints),
            itertools.repeat(args.nsteps), 
            itertools.repeat(args.beta_start),
            itertools.repeat(args.beta_end),
            itertools.repeat(args.stepsize_start),
            itertools.repeat(args.stepsize_end),
        ))
    best_result = min(results, key=lambda x: x.score)

    scores = np.array([result.scores for result in results])
    scores_mean = np.mean(scores, axis=0)
    scores_std = np.std(scores, axis=0)
    scores_min = np.min(scores, axis=0)
    scores_max = np.max(scores, axis=0)                
    score_results = np.array(
        [scores_mean, scores_std, scores_min, scores_max]
    )
    
    if args.score_file:
        write_score(
            args.score_file, score_results,
            header="avg(score) std(score) min(score) max(score)"
        )
    write_scheme(args.scheme_file, best_result, args.name)


    if args.show_space:
        figure = plt.figure(figsize=(FIGURE_WIDTH, FIGURE_WIDTH))
        ax = figure.gca()
        show_space(ax, space, lightness)
        figure.tight_layout()
    if args.show_scheme:
        figure = plt.figure(figsize=(FIGURE_WIDTH, FIGURE_WIDTH))
        ax = figure.gca()
        show_scheme(ax, space, best_result, lightness)
        figure.tight_layout()
    if args.show_example:
        # Check whether a custom non-amino-acid alphabet is used
        if args.alphabet is not None:
            raise InputError(
                "The example alignment can only be shown "
                "for the amino acid alphabet"
            )
        figure = plt.figure(figsize=(FIGURE_WIDTH, 2.5))
        ax = figure.gca()
        show_example(ax, best_result.rgb_colors)
        figure.tight_layout()
    if args.show_score:
        figure = plt.figure(figsize=(FIGURE_WIDTH, 6.0))
        ax = figure.gca()
        show_score(ax, best_result.scores)
        figure.tight_layout()
    if show_plots:
        plt.show()

    # In case someone wants to use the CLI results in a Python script
    # In any other case: Ignore the following two lines
    if result_container is not None:
        result_container.append(best_result)


def parse_alphabet(alphabet_str):
    if alphabet_str is None:
        return seq.LetterAlphabet(
            seq.ProteinSequence.alphabet.get_symbols()[:20]
        )
    else:
        if " " in alphabet_str:
            raise InputError("Alphabet may not contain whitespaces")
        try:
            return seq.LetterAlphabet(alphabet_str)
        except Exception:
            raise InputError("Invalid alphabet")


def parse_matrix(matrix_str, alphabet):
    if isfile(matrix_str):
        with open(matrix_str) as f:
            matrix_dict = align.SubstitutionMatrix.dict_from_str(f.read())
            return align.SubstitutionMatrix(alphabet, alphabet, matrix_dict)
    else:
        # String is a NCBI matrix name
        # For user convenience there is no case sensitivity
        # -> Find fitting matrix
        matrix_list = align.SubstitutionMatrix.list_db()
        upper_matrix_str = matrix_str.upper()
        upper_matrix_list = [
            m.upper() for m in align.SubstitutionMatrix.list_db()
        ]
        try:
            matrix_str = matrix_list[upper_matrix_list.index(upper_matrix_str)]
        except:
            raise InputError(
                f"'{matrix_str}' is neither a file "
                f"nor a valid NCBI substitution matrix"
            )
        return align.SubstitutionMatrix(alphabet, alphabet, matrix_str)


def adjust_saturation(space, smin, smax):
    lab = space.lab
    a = lab[..., 1]
    b = lab[... ,2]
    if smin is not None:
        space.remove(a**2 + b**2 < smin**2)
    if smax is not None:
        space.remove(a**2 + b**2 > smax**2)

def adjust_l(space, lmin, lmax):
    lab = space.lab
    l = lab[..., 0]
    if lmin is not None:
        space.remove(l < lmin)
    if lmax is not None:
        space.remove(l > lmax)

def adjust_a(space, amin, amax):
    lab = space.lab
    a = lab[..., 1]
    if amin is not None:
        space.remove(a < amin)
    if amax is not None:
        space.remove(a > amax)

def adjust_b(space, bmin, bmax):
    lab = space.lab
    b = lab[..., 2]
    if bmin is not None:
        space.remove(b < bmin)
    if bmax is not None:
        space.remove(b > bmax)


def write_scheme(file, result, name):
    write_color_scheme(file, result, name)

def write_score(file, results, header=""):    
    results = np.transpose(results)
    np.savetxt(file, results, fmt="%.2f", header=header)


def show_space(ax, space, lightness):
    # For performance reasons, filter correct lightness
    # before converting to RGB 
    l = space.lab[..., 0]
    space = copy.deepcopy(space)
    space.remove(l != lightness)
    # Remove first dimension (length = 1)
    rgb_space = space.get_rgb_space()[lightness]
    rgb_space[np.isnan(rgb_space)] = 0.7
    ax.imshow(np.transpose(rgb_space, axes=(1,0,2)), origin="lower",
                extent=(-128, 127,-128, 127), aspect="equal")
    ax.set_xlabel("a*")
    ax.set_ylabel("b*")
    ax.set_title(f"L* = {lightness}")

def show_scheme(ax, space, result, lightness):
    s = space.space
    ax.matshow(space.space[lightness].T, extent=(-128, 127,-128, 127),
               origin="lower", cmap=ListedColormap([(0.7,0.7,0.7), (1,1,1)]))
    for symbol, pos, color in zip(
        result.alphabet, result.lab_colors, result.rgb_colors
    ):
        ax.text(pos[1], pos[2], symbol, color=color,
                ha="center", va="center", size=14, weight="heavy")
    ax.xaxis.set_ticks_position("bottom")
    ax.set_xlabel("a*")
    ax.set_ylabel("b*")
    ax.set_title(f"L* = {lightness}")

def show_example(ax, colors):
    fasta_file = fasta.FastaFile.read(EXAMPLE_FILE_NAME)
    alignment = fasta.get_alignment(fasta_file)
    alignment = alignment[:60]

    graphics.plot_alignment_type_based(
        ax, alignment, spacing=2.0, symbols_per_line=len(alignment),
        color_scheme=colors
    )

def show_score(ax, scores):
    ax.plot(scores)
    ax.set_xlabel("Step")
    ax.set_ylabel("Score")


# This source code is part of the Gecos package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__author__ = "Patrick Kunzmann, Benjamin Mayer"
__all__ = ["ColorOptimizer", "ScoreFunction", "DefaultScoreFunction"]

from collections import namedtuple
import abc
import copy
import skimage
import numpy as np
import numpy.random as random
from .colors import lab_to_rgb


MIN_COORD = np.array([ 0, -128, -128], dtype=float)
MAX_COORD = np.array([99,  127,  127], dtype=float)


class ColorOptimizer(object):
    """
    Create an optimizer that tries to find an optimal color conformation
    within a given color space based on a score function.

    The optimizer tries to minimize the return value of the score
    function by adjusting the *Lab* values (coordinates) for each
    symbol in a given alphabet.

    The optimizer uses the random number generator from *NumPy*.
    Therefore, call :func:`numpy.random.seed()` to set the seed for the
    optimizer

    Parameters
    ----------
    alphabet : biotite.sequence.Alphabet
        The alphabet to calculate the color conformation for.
    score_function : ScoreFunction or callable
        The score function which should be minimized.
        When calling the object, its only parameter must be an array of
        coordinates with shape *(n, 3)*, where *n* is the length of the
        alphabet.
        Its return value must be a single float - the score.
    space : ColorSpace
        The color space that defines the allowed space for the
        coordinates.
    constraints : ndarray, shape=(n,3), dtype=float, optional
        An array whose non-NaN values are interpreted as constraints.
        Constrained values will be fixed during the optimization.
    """

    class Result(namedtuple("Result", ["alphabet", "trajectory", "scores"])):
        """
        The result of an optimization.
        Contains the final color scheme information as well as the
        course of the coordinates and the score during the optimization.

        Parameters
        ----------
        alphabet : biotite.sequence.Alphabet
            The alphabet the optimizer used.
        trajectory : ndarray, shape=(m,n,3), dtype=float
            The course of the coordinates during the simulation.
        scores : ndarray, shape=(m,), dtype=float
            The course of the score during the simulation.

        Attributes
        ----------
        alphabet : biotite.sequence.Alphabet
            The alphabet the optimizer used.
        trajectory : ndarray, shape=(m,n,3), dtype=float
            The course of coordinates during the simulation.
        lab_colors : ndarray, shape=(n,3), dtype=float
            The final *Lab* color conformation, i.e. the last element of
            `trajectory`.
        rgb_colors : ndarray, shape=(n,3), dtype=float
            The final color conformation converted into *RGB* colors.
        scores : ndarray, shape=(m,), dtype=float
            The course of the score during the simulation.
        score : float
            The final score, i.e. the last element of `scores`.
        
        """
        
        @property
        def score(self):
            return self.scores[-1]

        @property
        def lab_colors(self):
            return copy.deepcopy(self.trajectory[-1])
        
        @property
        def rgb_colors(self):
            return lab_to_rgb(np.floor(self.lab_colors))
    
    def __init__(self, alphabet, score_function, space, constraints=None):
        self._alphabet = alphabet
        self._n_symbols = len(alphabet)
        self._score_func = score_function
        self._space = space.space.copy()
        self._trajectory = []
        self._scores = []

        if constraints is None:
            self._constraint_mask = np.zeros(self._n_symbols, dtype=bool)
            self._constraints = np.full((self._n_symbols, 3), np.nan)
        else:
            self._constraint_mask = ~np.isnan(constraints).any(axis=-1)
            # Check constraints
            constraint_vals = constraints[self._constraint_mask]
            invalid_ind = np.where(~self._is_allowed(constraint_vals))[0]
            if len(invalid_ind) > 0:
                raise ValueError(
                    f"Constraint {constraint_vals[invalid_ind[0]]} "
                    f"is outside the allowed space"
                )
            self._constraints = constraints.copy()

        # Sample random initial coordinates
        coord = np.zeros((self._n_symbols, 3))
        self._apply_constraints(coord)
        self._set_coordinates(
            self._sample_coord(
                coord,
                lambda c: (
                    random.rand(*c.shape)
                    # Bring random values into correct range
                    * (MAX_COORD - MIN_COORD) + MIN_COORD
                )
            )
        )

    def set_coordinates(self, coord):
        """
        Set the the coordinates of the current color conformation.
        Potential color constraints are applied on these.
        This coordinate changes will be tracked in the trajectory.
        
        Parameters
        ----------
        coord : ndarray, shape=(n,3), dtype=float
            The new coordinates.
        """
        if coord.shape != (self._n_symbols, 3):
            raise ValueError(
                f"Given shape is {coord.shape}, "
                f"but expected shape is {(len(self._alphabet), 3)}"
            )

        invalid_ind = np.where(~self._is_allowed(coord))[0]
        if len(invalid_ind) > 0:
            raise ValueError(
                f"Coordinate {coord[invalid_ind[0]]} "
                f"is outside the allowed space"
            )

        coord = coord.copy()
        self._apply_constraints(coord)
        self._set_coordinates(coord)
        
    def _set_coordinates(self, coord, score=None):
        self._coord = coord
        self._trajectory.append(coord)
        if score is None:
            score = self._score_func(coord)
        self._scores.append(score)
    
    def optimize(self, n_steps=20000,
                 beta_start=1, beta_end=500,
                 stepsize_start=10, stepsize_end=0.2):
        r"""
        Perform a simulated annealing optimization on the current
        coordinate to minimize the score returned by the score function.
        
        This is basically a Metropolis-Monte-Carlo optimization where
        the inverse temperature and step size is varied according to
        exponential annealing schedule over the course of the
        optimization.                            
        
        Parameters
        ----------
        n_steps : int
            The number of simulated annealing steps.
        beta_start, beta_end : float
            The inverse temperature in the first and last step of the
            optimization, respectively.
            Higher values allow less increase of score, i.e. result
            in a steering into the local minimum.
            Must be positive.
        stepsize_start, stepsize_end : float
            The step size in the first and last step of the
            optimization, respectively.
            it is the radius in which the coordinates are randomly
            altered at in each optimization step.
            Must be positive.

        Notes
        -----
        The algorithm is a heuristic thats motivated by the physical
        process of annealing.
        If we, e.g., cool steel than a slow cooling can yield a superior
        quality, whereas for a fast cooling the steel can become
        brittle.
        The same happens here within the search space for the given
        minimization task.       
        """
        betas = _calculate_schedule(n_steps, beta_start, beta_end)
        stepsizes = _calculate_schedule(n_steps, stepsize_start, stepsize_end)

        for i in range(n_steps):
            score = self._scores[-1]
            new_coord = self._sample_coord(
                self._coord,
                lambda c: c + (random.rand(*c.shape)-0.5) * 2 * stepsizes[i]
            )
            new_score = self._score_func(new_coord)
            
            if new_score < score:
                self._set_coordinates(new_coord, new_score)
                
            else:
                p_accept = np.exp( -betas[i] * (new_score-score))
                if random.rand() <= p_accept:
                    self._set_coordinates(new_coord, new_score)
                else:
                    self._set_coordinates(self._coord, score)
                    
    def get_result(self):
        """
        Get the result of the optimization.

        Returns
        -------
        result : ColorOptimizer.Result
            The result.
        """
        trajectory = np.array(self._trajectory)
        scores = np.array(self._scores)
        return ColorOptimizer.Result(
            alphabet = self._alphabet,
            trajectory = trajectory,
            scores = scores
        )
    
    def _is_allowed(self, coord):
        """
        Get a mask indicating allowed color values, i.e. values within
        the color space.
        """
        mask = ((coord >= MIN_COORD) & (coord <= MAX_COORD)).all(axis=-1)
        # Only check values that are within valid index range
        # Use floor for correct conversion of negative values to index
        ind = np.floor(coord[mask] - MIN_COORD).astype(int)
        mask[mask.copy()] = self._space[ind[..., 0], ind[..., 1], ind[..., 2]]
        return mask
    
    def _sample_coord(self, old_coord, sampler):
        """
        Based on given coordinates sample new coordinates that are
        within the color space.
        The sampling function must accept a given array of coordinates
        and return a new array of the same size.
        """
        new_coord = old_coord.copy()
        # Resample coordinates that are not in valid space until they
        # are in valid space
        resample_mask = ~self._constraint_mask.copy()
        while resample_mask.any():
            new_coord[resample_mask] = sampler(old_coord[resample_mask])
            resample_mask[self._is_allowed(new_coord)] = False
        return new_coord

    
    def _apply_constraints(self, coord):
        coord[self._constraint_mask] = self._constraints[self._constraint_mask]


def _calculate_schedule(n_steps, start, end):
    """
    Calculate the values for each step in an exponential schedule.
    """
    # Use float 64
    return start * (end/start)**np.linspace(0, 1, n_steps, dtype=np.float64)



class ScoreFunction(metaclass=abc.ABCMeta):
    """
    Abstract base class for a score function.
    A score function calculates a score from a color conformation
    (coordinates).

    The score is calculated by calling the object with the coordinates
    as single argument.
    Hence, classes inheriting from this base class mut override the
    :func:`__call__()` method.

    Parameters
    ----------
    n_symbols : int
        The amount of symbols in the system.
        Equivalent to the length of the alphabet the color scheme is
        generated for.
        This value is used to check the shape of the coordinates when
        calling the score function.
    """

    def __init__(self, n_symbols):
        self._n_symbols = n_symbols

    @abc.abstractmethod
    def __call__(self, coord):
        """
        Calculate the score for the given coordinates.

        Parameters
        ----------
        coord : ndarray, shape=(n,3), dtype=float
            The coordinates.
        
        Returns
        -------
        score : float
            The score assigned to `coord`.
        """
        if len(coord) != self._n_symbols:
            raise ValueError(
                f"Expected {self._n_symbols} coordinates, but got {len(coord)}"
            )


class DefaultScoreFunction(ScoreFunction):
    """
    Create an instance of the default score function *Gecos* uses.

    The score function contains two terms:
    A sum of harmonic potentials between each pair of symbols, based on
    a substitution matrix, and *contrast score* that favors schemes with
    a high contrast.

    Parameters
    ----------
    matrix : biotite.sequence.align.SubstitutionMatrix
        A distance matrix is calculated from this score matrix.
        The equilibrium positions scale linearly with the values in the
        distance matrix.
    contrast : int, optional
        A weight for the *contrast score*.
    distance_formula : {'CIE76', 'CIEDE94', 'CIEDE2000'}, optional
        The formula to use for calculation of perceptual color
        difference.
        While ``'CIEDE2000'`` is the most accurate formula for the
        perceptual difference, ``'CIE76'`` features the fastest
        calculation.
    """
    def __init__(self, matrix, contrast=700, distance_formula="CIEDE2000"):
        if not matrix.is_symmetric():
            raise ValueError("Substitution matrix must be symmetric")
        n_symbols = len(matrix.get_alphabet1())
        super().__init__(n_symbols)
        self._tri_indices = np.tril_indices(n_symbols, k=-1)
        self._ideal_dist = DefaultScoreFunction._calculate_ideal_distances(
            self._tri_indices, matrix
        )
        self._contrast = contrast
        if distance_formula not in ["CIE76", "CIEDE94", "CIEDE2000"]:
            raise ValueError(
                f"Unknown color distance formula '{distance_formula}'"
            )
        self._distance_formula = distance_formula
    
    def __call__(self, coord):
        super().__call__(coord)
        
        dist = DefaultScoreFunction._calculate_distances(
            self._tri_indices, coord, self._distance_formula
        )
        # This factor translates visual distances
        # into normalized substitution matrix distances
        mean_dist = np.average(dist)
        # Harmonic potentials between each pair of symbols
        harmonic_loss = np.sum((dist / mean_dist - self._ideal_dist)**2)
        # Contrast term: Favours conformations
        # with large absolute color differences
        contrast_loss = self._contrast / mean_dist
        return harmonic_loss + contrast_loss
    
    @staticmethod
    def _calculate_distances(tri_indices, coord, distance_formula):
        ind1, ind2 = tri_indices
        if distance_formula == "CIE76":
            return np.sqrt(
                np.sum((coord[ind1, :] - coord[ind2, :])**2, axis=-1)
            )
        elif distance_formula == "CIEDE94":
            return skimage.color.deltaE_ciede94(
                coord[ind1], coord[ind2]
            )
        else: #"CIEDE2000"
            return skimage.color.deltaE_ciede2000(
                coord[ind1], coord[ind2]
            )

    @staticmethod
    def _calculate_ideal_distances(tri_indices, substitution_matrix):
        scores = substitution_matrix.score_matrix()
        diff_to_max = np.diag(scores) - scores
        dist_matrix = (diff_to_max + diff_to_max.T) / 2
        ind_i, ind_j = tri_indices
        distances = dist_matrix[ind_i, ind_j]
        # Scale, so that average distance is 1
        distances /= np.average(distances)
        return distances



# This source code is part of the Gecos package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__author__ = "Patrick Kunzmann"
__all__ = ["ColorSpace"]

from os.path import join, dirname, realpath
import itertools
import numpy as np
from .colors import lab_to_rgb


SPACE_FILE_NAME = join(dirname(realpath(__file__)), "space.npy")


class ColorSpace():
    """
    Create a color space, that spans the complete *RGB* gamut of the *Lab*
    color space.
    The *Lab* components are discretized into integer values.

    A color space describes the boundaries of the color scheme to be generated.
    It uses the *Lab* color space.
    In addition to the inherent limit to colors, that can also be displayed in
    *RGB* space, further parts of the space can be removed by calling
    :func:`remove()`.

    Parameters
    ----------
    file_name : str, optional
        The path of a custom file to load the precalculated *RGB* convertible
        space from.

    Attributes
    ----------
    shape : tuple of int
        The shape of the space, i.e. the amount of *Lab* values in each
        dimension.
    lab : ndarray, shape=(100, 256, 256, 3), dtype=int
        The complete discretized *Lab* color space in the *ab* range of
        ``-128`` to ``127``.
    space : shape=(100, 256, 256), dtype=bool
        The allowed part of the `lab` attribute, i.e. the part that is
        convertible into *RGB* and was not manually removed.
    """

    def __init__(self, file_name=None):
        if file_name is None:
            file_name = SPACE_FILE_NAME
        
        l = np.arange(100)
        a = b = np.arange(-128,128)
        self._lab = np.zeros((100,256,256,3), dtype=int)
        self._lab[:,:,:,0] = l[:, np.newaxis, np.newaxis]
        self._lab[:,:,:,1] = a[np.newaxis, :, np.newaxis]
        self._lab[:,:,:,2] = b[np.newaxis, np.newaxis, :]

        with open(file_name, "rb") as file:
            self._space = np.load(file)

    def remove(self, mask):
        """
        Remove a portion of the color space.
        
        Parameters
        ----------
        space : ndarray, shape=(100, 256, 256), dtype=bool
            The space is removed where this mask is true.
        
        Examples
        --------
        Remove space below a defined lightness:

        >>> L_MIN = 50
        >>> space = ColorSpace()
        >>> lab = space.lab
        >>> l = lab[..., 0]
        >>> space.remove(l < L_MIN)
        """
        self._space &= ~mask
    
    def get_rgb_space(self):
        """
        Convert the *Lab* colors of the space in *RGB* colors.
        
        Returns
        -------
        rgb : ndarray, shape=(100,256,256,3), dtype=float
            The *RGB* colors.
            Colors that cannot be displayed in *RGB* or were manually
            removed from the space are ``NaN``.
        """
        rgb = lab_to_rgb(self._lab)
        rgb[~self._space] = np.nan
        return rgb

    @property
    def space(self):
        return self._space.copy()

    @property
    def shape(self):
        return (100, 256, 256)
    
    @property
    def lab(self):
        return self._lab.copy()

    @staticmethod
    def _generate(file_name=None):
        """
        Precalculate which part of the *Lab* space can be displayed in
        *RGB* and save the result as boolean mask into a
        *NumPy* ``.npy`` file.
        """
        lab = np.zeros((100, 256, 256, 3), dtype=int)
        lab[:,:,:,0] = np.arange(100      )[:, np.newaxis, np.newaxis]
        lab[:,:,:,1] = np.arange(-128, 128)[np.newaxis, :, np.newaxis]
        lab[:,:,:,2] = np.arange(-128, 128)[np.newaxis, np.newaxis, :]
        
        rgb = lab_to_rgb(lab)
        space = np.ones((100, 256, 256), dtype=bool)
        space[np.isnan(rgb).any(axis=-1)] = False
        
        if file_name is None:
            file_name = SPACE_FILE_NAME
        with open(file_name, "wb") as file:
            np.save(file, space)


# This source code is part of the Gecos package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__author__ = "Patrick Kunzmann"
__all__ = ["write_color_scheme"]

import json
from matplotlib.colors import to_hex


def write_color_scheme(file, result, name=""):
    """
    Write a color scheme in the *Biotite* compatible JSON format into
    a file.
    
    Parameters
    ----------
    file : file-like object
        The file to write the scheme into.
    result : Optimizer.Result
        The result from the optimization.
        Contains the color scheme.
    name : str, optional
        Name of the scheme. Will be written to the value of the
        ``"name"`` key.
    """
    scheme = {}
    scheme["name"] = name
    symbols = result.alphabet.get_symbols()
    scheme["alphabet"] = symbols
    scheme["colors"] = {s : to_hex(c) for s, c
                        in zip(symbols, result.rgb_colors)}
    json.dump(scheme, file, indent=4)


# This source code is part of the Gecos package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__version__ = "2.0.0"
__author__ = "Patrick Kunzmann"

from .colors import *
from .space import *
from .optimizer import *
from .file import *



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['gecos']

package_data = \
{'': ['*']}

install_requires = \
['biotite>=0.21',
 'matplotlib>=3.0,<4.0',
 'numpy>=1.13,<2.0',
 'scikit-image>=0.17']

entry_points = \
{'console_scripts': ['gecos = gecos.cli:main']}

setup_kwargs = {
    'name': 'gecos',
    'version': '2.0.0',
    'description': 'Generated color schemes for sequence alignment visualizations',
    'long_description': 'Gecos - Generated Color Schemes for sequence alignment visualizations\n=====================================================================\n\nMultiple sequence alignments are often visualized by coloring the symbols\naccording to some kind of properties.\nFor example a color scheme for amino acids could use one color for\nhydrophobic residues, another color for positively charged residues\nand so forth.\nUsually, such color schemes are manually created by experienced people\nwho have knowledge about the characteristics of the e.g. amino acids,\nso they can assign equal or similar colors to amino acids that share\nsimilar properties.\n\nThe *Gecos* software follows a different approach:\nInstead of looking at specific, sometimes subjective properties,\nit uses another source for estimating the similarity of symbols:\nthe substitution matrix itself.\nSimilar colors are assigned to high scoring pairs of symbols, low\nscoring pairs get distant colors - in a completely automatic manner.\nAs a result the distance of two symbols in the substitution matrix corresponds\nto the perceptual differences in the color scheme.\n\nHow about an example?\nThe following command line invocation creates a light color scheme.\nAn example alignment using the newly generated color scheme is displayed below.\n\n.. code-block:: console\n   \n   $ gecos --matrix BLOSUM62 --lmin 60 --lmax 75 -f awesome_colors.json\n\n.. image:: https://raw.githubusercontent.com/biotite-dev/gecos/master/doc/static/assets/figures/main_example_alignment.png\n\nInstallation\n============\n\nIn order to use *Gecos* you need to have Python (at least 3.6) installed.\nFurthermore, the following Python packages are required:\n\n   - **biotite**\n   - **numpy**\n   - **matplotlib**\n   - **scikit-image**\n\nIf these prerequisites are met, *Gecos* is simply installed via\n\n.. code-block:: console\n\n   $ pip install gecos\n\nAlternatively, *Gecos* can be installed via *Conda*:\n\n.. code-block:: console\n\n   $ conda install -c conda-forge gecos\n\nCitation\n--------\n\nIf you use *Gecos* in a scientific publication, please cite:\n\n|\n\n   P. Kunzmann, B. E. Mayer, K. Hamacher,\n   "Substitution matrix based color schemes for sequence alignment visualization,"\n   BMC Bioinformatics, vol. 21, pp. 209, 2020.\n   doi: `10.1186/s12859-020-3526-6 <https://doi.org/10.1186/s12859-020-3526-6>`_',
    'author': 'Patrick Kunzmann',
    'author_email': 'patrick.kunzm@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gecos.biotite-python.org',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
